@@ -8,28 +8,32 @@ typedef Display* DisplayOrNull; /* Used by typemap for stricter conversion */
 typedef Visual* VisualOrNull;
 typedef int ScreenNumber; /* used by typemap to coerce X11::Xlib::Screen */
 
-/* Methods to create/alter the magic Display* attached to X11::Xlib objects */
+/* Functions to create/alter the magic Display* attached to X11::Xlib objects */
 
 extern Display * PerlXlib_get_magic_dpy(SV *sv, Bool not_null);
 extern SV * PerlXlib_set_magic_dpy(SV *sv, Display *dpy);
 extern SV * PerlXlib_obj_for_display(Display *dpy, int create);
 
+/* un-pack an XID from a wrapped X11::Xlib::XID or subclass */
 extern XID PerlXlib_sv_to_xid(SV *sv);
 
-extern int PerlXlib_keysym_to_codepoint(KeySym keysym);
-extern KeySym PerlXlib_codepoint_to_keysym(int codepoint);
-extern SV * PerlXlib_keysym_to_sv(KeySym keysym, int symbolic);
-extern KeySym PerlXlib_sv_to_keysym(SV *sv);
+/* Functions to wrap/unwrap opaque X11 pointers to/from objects */
+extern void * PerlXlib_sv_to_display_innerptr(SV *sv, bool not_null);
+extern SV * PerlXlib_obj_for_display_innerptr(Display *dpy, void *thing, const char *thing_class, int svtype, bool create);
+extern void * PerlXlib_get_magic_dpy_innerptr(SV *sv, Bool not_null);
+extern SV * PerlXlib_set_magic_dpy_innerptr(SV *sv, void *innerptr);
+/* but Screen* is special */
+extern Screen * PerlXlib_sv_to_screen(SV *sv, bool not_null);
+extern SV * PerlXlib_obj_for_screen(Screen *screen);
 
+/* generically attach Display to any pointer-based object */
+extern SV * PerlXlib_get_displayobj_of_opaque(void *thing);
+extern void PerlXlib_set_displayobj_of_opaque(void *thing, SV *dpy_sv);
+
+/* Functions to pack/unpack structs into blessed scalars */
 typedef void PerlXlib_struct_pack_fn(void*, HV*, Bool consume);
-
 extern void* PerlXlib_get_struct_ptr(SV *sv, int lvalue, const char* pkg, int struct_size, PerlXlib_struct_pack_fn *packer);
-extern void PerlXlib_install_error_handlers(Bool nonfatal, Bool fatal);
-
 extern const char* PerlXlib_xevent_pkg_for_type(int type);
-
-/* Pack and Unpack functions for structs */
-
 extern void PerlXlib_XEvent_pack(XEvent *s, HV *fields, Bool consume);
 extern void PerlXlib_XEvent_unpack(XEvent *s, HV *fields);
 extern void PerlXlib_XVisualInfo_pack(XVisualInfo *s, HV *fields, Bool consume);
@@ -44,3 +48,15 @@ extern void PerlXlib_XSizeHints_pack(XSizeHints *s, HV *fields, Bool consume);
 extern void PerlXlib_XSizeHints_unpack(XSizeHints *s, HV *fields);
 extern void PerlXlib_XRectangle_pack(XRectangle *s, HV *fields, Bool consume);
 extern void PerlXlib_XRectangle_unpack(XRectangle *s, HV *fields);
+#ifdef HAVE_XRENDER
+extern void PerlXlib_XRenderPictFormat_pack(XRenderPictFormat *s, HV *fields, Bool consume);
+extern void PerlXlib_XRenderPictFormat_unpack(XRenderPictFormat *s, HV *fields);
+#endif
+
+/* Keysym/unicode utility functions */
+extern int PerlXlib_keysym_to_codepoint(KeySym keysym);
+extern KeySym PerlXlib_codepoint_to_keysym(int codepoint);
+extern SV * PerlXlib_keysym_to_sv(KeySym keysym, int symbolic);
+extern KeySym PerlXlib_sv_to_keysym(SV *sv);
+
+extern void PerlXlib_install_error_handlers(Bool nonfatal, Bool fatal);
